# 监视器ubuntu20.04 网络配置 bsymv
# last modify by cls on 20210830
# 查找proc剔除无线网卡
import sys
import time
import re
import os
import json
import math
import ctypes
from ctypes import *
import time
from threading import Timer
import fcntl
import subprocess
import select

#3.0.1 增加了服务版 wifi设置 但需要重启后才能生效
VERSION_SHOW = ("version 3.0.2 for ubuntu2004-moju on %s") % time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(time.time()))

workpath = os.path.dirname(os.path.realpath(__file__))    # 从当前文件路径中获取目录

netplan_yaml = "/etc/netplan/default_netplan.yaml"  # netplan配置文件 yaml
netplan_config_path = "/etc/netplan/netplan_prejson.txt"
netplan_config_path_wifi = "/etc/netplan/wifi_prejson.txt"  #wifi配置
# c/c++动态库
#curpath = os.getcwd()
clib_clsnt_path = workpath + "/" + "libcls_netclib.so"  # cls network tool

#clib_clsnt_path = "/root/cls_netclib/" + "libcls_netclib.so"  # cls network tool

# 解析json 配置文件内容netplan_yaml = "/etc/netplan/default_netplan.yaml"  # netplan配置文件 yaml

# netplan_yaml = "/root/bin-AOI/cnt_net/netplan_preyaml.yaml"
# netplan_config_path = "/root/bin-AOI/cnt_net/netplan_prejson.txt"  # 配置文件json（自定义）

# function---------------------------------------------------------------------------
# 删除转义反斜杠
def removequote(str: str):
    str = str.replace("\\\\", "\\")
    # str = str.replace("\\\"", "\"")
    return str


# 掩码ip转掩码长度
def mask2masklen(maksip: str):
    retback = 24  # 如果解析错误 默认返回值24
    sppart = maksip.split(".")
    if len(sppart) != 4:
        return retback
    retback = 0
    for item in sppart:
        i = int(item)
        str2b = format(i, "b")  #将10进制 转成 2进制 字符串
        retback += str2b.count('1')
    return retback

#根据ip和mask判断所属子网地址
def getIpPart(ip: str, mask: str):
    ippart = ip.split(".")
    maskpart = mask.split(".")
    if len(ippart) != 4 or len(maskpart) != 4:
        return None
    finalpart = []
    for i in range(4):
        iip = int(ippart[i])
        imask = int(maskpart[i])
        p_i = iip & imask   #算分类
        finalpart.append(str(p_i))
    return ".".join(finalpart)

def isWifi(name: str):
    path = "/sys/class/net/%s/wireless" % name
    if os.path.exists(path):
        return True
    return False

class CCommand:
    TIMEOUT_MSG = ""

    def __init__(self, cmd):
        self.runtime = 0
        self.cmd = cmd

    # 执行函数 默认20秒超时
    def make_nonblock(self, fd):
        flags = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)

    #启动
    def command_run(self, timeout=5):
        self.proc = subprocess.Popen(
            self.cmd, stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        self.make_nonblock(self.proc.stdout)
        self.make_nonblock(self.proc.stderr)
        self.runtime = time.time()
        timer = None
        ret = {}
        out_ret = ""
        err_ret = ""
        if timeout:
            timer = Timer(timeout, self._kill_proc, [self.proc, ])
            timer.setDaemon(True)
            timer.start()
        while True:
            out_ret += self._read(self.proc.stdout)
            err_ret += self._read(self.proc.stderr)
            if self.proc.poll() is not None:  # 不在运行状态
                break
        out_ret += self._read(self.proc.stdout)
        err_ret += self._read(self.proc.stderr)
        self.runtime = time.time() - self.runtime  # time
        ret["OUT"] = out_ret
        ret["ERR"] = self.TIMEOUT_MSG + err_ret
        ret["STATUS"] = self.proc.returncode
        ret["RUNTIME"] = "%.3f s" % self.runtime
        if timer:
            timer.cancel()
            timer.join()
        self.proc.stdout.close()
        return ret

    def _read(self, fd):
        out = ""
        if fd is None or fd.closed:
            return out
        rd, _, _ = select.select([fd], [], [], 0)
        if rd:
            outs = fd.readlines()
            for itemout in outs:
                out += itemout.decode(encoding='utf-8')
        return out

    def _kill_proc(self, proc):
        if proc:
            try:
                proc.kill()
                self.TIMEOUT_MSG = "timeout:"
            except OSError:
                pass
            finally:
                proc.stdout.close()

# 此程序函数---------------------------------------------------------------------------
# 检查并构建 新的node内容, listnode接修改后的结构组
# 判断重复的网段
# 检查使用唯一的网关
def reBuildNodes(listnode: dict):
    worknodes = list(listnode.keys())

    # 检查是否有注册的网关信息, 有写网关的网卡优先使用ip, 不判断修改
    register_route = chkbuildNodes_findroute(worknodes, listnode)
    if register_route and register_route in worknodes:
        worknodes.remove(register_route)

    # 全部按静态处理剩余项
    chkbuildNodes_repetition(worknodes, listnode)

# 找到dhcp 优先设置
def chkbuildNodes_findroute(worknodes: list, dictchk: dict) -> str:
    register_route = ""  # 需要使用的网关信息 ->mac
    # 先找dhcp设置的项
    for inode in dictchk.values():
        if inode.mac not in worknodes:
            continue
        if inode.isdhcp:
            inode.status = "normal"
            inode.describe = "使用dhcp获取动态ip"
            inode.address = ""
            inode.netmask = ""
            inode.gateway = ""
            inode.dns = ""
            register_route = inode.mac
            break
    if register_route:
        return register_route
    for inode in dictchk.values():
        if inode.mac not in worknodes:
            continue
        if inode.gateway:  # 如果有设置gateway, 判断ip是否是同一网段
            # [('192.168.5.5', '192.168.5')]
            ippart = getIpPart(inode.address, inode.netmask)
            routepart = getIpPart(inode.gateway, inode.netmask)
            if not ippart or not routepart or ippart != routepart :
                continue
            inode.status = "normal"
            inode.describe = "优先使用的网关"
            inode.isdhcp = False
            if not inode.dns:
                inode.dns = inode.gateway
            register_route = inode.mac
            break
    return register_route

#判定静态ip配置 是否有重复的ip
def chkbuildNodes_repetition(worknodes: list, dictchk: dict):
    listregister = []

    # 跳过静态ip设置的项
    routeitem = list(set(worknodes) ^ dictchk.keys())
    if routeitem and routeitem[0] in dictchk.keys():
        useipbyroute = dictchk[routeitem[0]].address
        search_ip = re.findall("^((\d+.\d+.)(\d+).\d+)", useipbyroute, re.M)
        if search_ip:
            part1 = search_ip[0][1]
            ipart2 = int(search_ip[0][2])
            opart = part1 + "%d" % ipart2
            listregister.append(opart)

    for inode in dictchk.values():
        if inode.mac not in worknodes:
            continue
        # 判断ip是否重复
        ip = inode.address
        if not ip:
            ip = "192.168.3.5"
        search_ip = re.findall("^((\d+.\d+.)(\d+).\d+)", ip, re.M)
        if not search_ip:
            continue
        # [('192.168.1.77',  '192.168.', '1')]
        aip = search_ip[0][0]
        part1 = search_ip[0][1]
        ipart2 = int(search_ip[0][2])
        opart = part1 + "%d" % ipart2
        startipart2 = 3
        while True:
            if startipart2 > 253 or startipart2 < 3:
                break
            if opart not in listregister:  # 没有使用过的网段
                inode.address = aip
                listregister.append(opart)
                inode.describe = "设置成功"
                break
            else:
                startipart2 = startipart2 % 255 + 1
                opart = part1 + "%d" % startipart2
                inode.status = "warning"
                inode.describe = "存在相同的网段,已经更新至 %s" % opart
                aip = opart + ".5"
        inode.netmask = "255.255.255.0"
        inode.gateway = ""
        inode.dns = ""
        inode.isdhcp = False

# 判断netplan应用环境 桌面版：NetworkManager，服务版：networkd
def getNetplanContext() -> str:
    dcmd = CCommand("NetworkManager --version")
    iprouteshow = dcmd.command_run()
    back = iprouteshow["OUT"]
    workmanager = "NetworkManager"
    if back:
        workmanager = "NetworkManager"
    else:
        workmanager = "networkd"
    return workmanager

# 创建netplan规则文件
def buildNetPlanFile(listnode, listnode_wifi):
    workmanager = getNetplanContext()   #"NetworkManager" / "networkd"

    str = ""
    use_t = " " * 4  # 4空格
    str += "# This is the network config written by 'subiquity'" + "\n"
    str += "network:" + "\n"
    str += use_t + "ethernets:" + "\n"
    for node in listnode.values():
        str += node.getYamlStr()
    #wifi
    if len(listnode_wifi):
        wifi_str = ""
        for node_wifi in listnode_wifi.values():
            wifi_str += node_wifi.getYamlStr()
        if wifi_str:
            str += use_t + "wifis:\n" + wifi_str
    str += use_t + "version: 2" + "\n"
    str += use_t + "renderer: %s \n" % workmanager

    os.system("rm -f /etc/netplan/*.yaml")
    with open(netplan_yaml, "w+") as yaml_f:
        yaml_f.write(str)

#更新状态
def finshUpdatePrejson(listnode : dict, prejson_filepath: str):
   # 等待网络配置plan 获取报表
    waitreporttime = time.time()
    afterplan_waitchk = listnode.keys()
    roundtimes = 0
    while roundtimes < 10 and afterplan_waitchk:
        savelist = []
        for item in listnode:
            itemobj = listnode[item]
            if not itemobj.use:
                continue
            if not itemobj.isdhcp and not itemobj.gateway and roundtimes > 4:
                continue
            back = itemobj.rechkafterplan(libc_clsnt)  # 检查计划执行结果
            if back:
                continue
            savelist.append(item)
        afterplan_waitchk = savelist
        roundtimes += 1
        if afterplan_waitchk:
            time.sleep(1)
    waitreporttime = time.time() - waitreporttime
    print("wait report time:%.3f s" % waitreporttime)

    # 保存新的配置文件
    #有线
    mapReport = {}
    report = []
    for item in listnode.values():
        report.append(item.getReport())
    mapReport["netcard_info"] = report
    jsonstr = json.dumps(mapReport, indent=4, separators=(',', ':'), ensure_ascii=False)
    jsonstr = removequote(jsonstr)
    with open(prejson_filepath, "w+") as f_report:
        f_report.write(jsonstr)

# class---------------------------------------------------------------------------
class ND_Node:
    def __init__(self):
        self.mac = ""
        self.name = ""
        self.address = ""
        self.netmask = "255.255.255.0"
        self.gateway = ""
        self.dns = ""
        self.isdhcp = False
        self.describe = ""
        self.describe_netplan = ""
        self.status = "normal"  # normal, warning, error
        self.iswifi = False
        self.use = True

    # 从配置文件中加载参数信息
    def uploadConfig(self):
        try:
            loaddict = {}
            with open(netplan_config_path, "r") as json_f:
                loaddict = json.load(json_f)
                jsonnodes = loaddict["netcard_info"]

            for inode in jsonnodes:
                imac = inode["mac"]
                if imac not in self.mac:
                    continue
                if inode["address"] and inode["address"] != "0.0.0.0":
                    self.address = inode["address"]
                if inode["netmask"] and inode["netmask"] != "0.0.0.0":
                    self.netmask = inode["netmask"]
                if inode["gateway"] and inode["gateway"] != "0.0.0.0":
                    self.gateway = inode["gateway"]
                if inode["dns"] and inode["dns"] != "0.0.0.0":
                    self.dns = inode["dns"]
                self.isdhcp = inode["isdhcp"]
        except KeyError as Argument:
            print("KeyError", Argument)
        except json.decoder.JSONDecodeError as Argument:
            print("JSONDecodeError", Argument)
            os.system("rm -f " + netplan_config_path)
        except FileNotFoundError as f:
            print("FileNotFoundError" , f)

    def getYamlStr(self):
        str = ""
        use_t = " " * 4 #4空格

        str += (use_t * 2) + self.name + ":" + "\n"
        if self.isdhcp:
            str += (use_t * 3) + "dhcp4: " + "true" + "\n"
            return str
        else:
            str += (use_t * 3) + "dhcp4: " + "no" + "\n"

        str += (use_t * 3) + "addresses: " + "[" + self.address + "/" + "%d" % mask2masklen(
            self.netmask) + "]" + "\n"
        if not isWifi(self.name):
            str += (use_t * 3) + "mtu: " + "%d" % 9000 + "\n"
        if self.gateway and self.gateway != "0.0.0.0":
            str += (use_t * 3) + "gateway4: " + self.gateway + "\n"
            str += (use_t * 3) + "nameservers: " + "\n"
            str += (use_t * 4) + "addresses: " + "[" + self.dns + "]" + "\n"
        return str

    def getReport(self):
        return self.__dict__

    def rechkafterplan(self, libc):
        class CSTU_INFO(Structure):
            _fields_ = [("ip", c_char * 20),
                        ("mask", c_char * 20)]

        obs_info = CSTU_INFO()  # object struct info
        ptr_stu = byref(obs_info)  # to -> ptr
        # func
        cfun_getinfo = libc.NC_getEnInfo
        cfun_getinfo.restype = c_bool
        # libnet.argtypes = [c_char_p, CSTU_INFO]
        name = create_string_buffer(self.name.encode("utf-8"))
        rback = cfun_getinfo(name, ptr_stu)

        curip = str(obs_info.ip, 'utf-8')
        curmask = str(obs_info.mask, 'utf-8')

        # print(rback, curip, curmask)

        if rback:
            self.describe_netplan = ""

            if self.gateway or self.isdhcp:
                route = ""
                dns = ""

                # route
                dcmd = CCommand("ip route show")
                iprouteshow = dcmd.command_run()
                back = iprouteshow["OUT"]
                reachroute = re.findall("^default via (\d+.\d+.\d+.\d+)\sdev\s(\S+)\sproto.*", back)
                if reachroute and reachroute[0][1] == self.name:
                    route = reachroute[0][0]
                # dns
                etc_resolv = ""
                with open("/etc/resolv.conf", "r") as f_resolv:
                    etc_resolv = f_resolv.read()
                reachdns = re.findall("nameserver (\d+.\d+.\d+.\d+)", etc_resolv)
                if reachdns:
                    dns = reachdns[0]

                if not route:
                    self.describe_netplan = "网关未配置成功"
                    self.status = "warning"
                if not dns or dns != "127.0.0.53":
                    self.describe_netplan = "DNS未配置成功"
                    self.status = "warning"
                self.address = curip
                self.netmask = curmask
                self.gateway = route
                #self.dns = dns
                self.describe_netplan = "获取成功,正常使用"

                # 检查是否有和其他网卡的网段冲突
                for otheritem in gMName4Node.keys():
                    if otheritem == self.mac:
                        continue
                    objitem = gMName4Node[otheritem]
                    reachpart = re.findall("(\d+.\d+.\d+).\d+", objitem.address)
                    reachpart_in = re.findall("(\d+.\d+.\d+).\d+", self.address)
                    if reachpart and reachpart_in:
                        if reachpart_in and reachpart[0] == reachpart_in[0]:
                            self.describe_netplan = "与%s网卡使用的网段冲突了" % objitem.name
                            self.status = "error"
            else:
                # 判断获取的和预设的是否一样
                if curip != self.address or curmask != self.netmask:
                    self.describe_netplan = "配置结果与预设值有偏差" + curip + "/" + curmask
                    self.status = "warning"
                else:
                    self.describe_netplan = "正常使用"

            return True
        else:
            self.describe_netplan = "网卡可能未使用"
            self.status = "warning"

        return False

class ND_Node_Wifi(ND_Node):
    def __init__(self):
        super().__init__()
        self.iswifi = True
        self.isdhcp = True
        self.use = False
        self.hotname = ""
        self.passwd = ""

    def uploadConfig(self):
        try:
            loaddict = {}
            with open(netplan_config_path_wifi, "r") as json_f:
                loaddict = json.load(json_f)
                jsonnodes = loaddict["netcard_info"]

            for inode in jsonnodes:
                imac = inode["mac"]
                if imac not in self.mac:
                    continue
                if inode["address"] and inode["address"] != "0.0.0.0":
                    self.address = inode["address"]
                if inode["netmask"] and inode["netmask"] != "0.0.0.0":
                    self.netmask = inode["netmask"]
                if inode["gateway"] and inode["gateway"] != "0.0.0.0":
                    self.gateway = inode["gateway"]
                if inode["dns"] and inode["dns"] != "0.0.0.0":
                    self.dns = inode["dns"]
                if inode["use"]:
                    self.use = inode["use"]
                if inode["hotname"]:
                    self.hotname = inode["hotname"]
                if inode["passwd"]:
                    self.passwd = inode["passwd"]
                self.isdhcp = inode["isdhcp"]
        except KeyError as Argument:
            print("KeyError", Argument)
        except json.decoder.JSONDecodeError as Argument:
            print("JSONDecodeError", Argument)
            os.system("rm -f " + netplan_config_path)
        except FileNotFoundError as Argument:
            print("FileNotFoundError", Argument)

    def getYamlStr(self):
        if not self.use:
            return ""
        if not self.hotname or not self.passwd:
            return ""
        use_t = " " * 4
        str = super().getYamlStr()
        str += (use_t * 3) + "access-points: \n"
        str += (use_t * 4) + "\"%s\": \n" % self.hotname
        str += (use_t * 5) + "password: \"%s\"\n" % self.passwd
        return str

# main---------------------------------------------------------------------------
if __name__ == "__main__":
    print(VERSION_SHOW)

    cdll.LoadLibrary(clib_clsnt_path)
    libc_clsnt = CDLL(clib_clsnt_path)

    os.system("mkdir -p ./cnt_net")
    gMName4Node = {} #普通网卡 # key:网卡名 value:对象
    gMName4Node_wifi = {}   #wifi信息

    # /proc/net/dev : listproc_cardname ['enp4s0', 'enp2s0', 'lo', 'enp3s0', 'enp6s0', 'enp5s0']
    # 获取mac和网卡名对应的字典 : dict_mac2name {'b0:41:6f:04:5a:5e': 'eno1', 'b0:41:6f:04:4b:d6': 'enp2s0'}
    dict_mac2name = {} #无wifi
    dict_mac2name_wifi = {} #wifi列表
    f = open("/proc/net/dev", "r")
    str_cmd = f.read()
    f.close()
    listproc_cardname = re.findall("\s+(\S+):", str_cmd)    #所有发现的网卡
    for iproc in listproc_cardname:
        if iproc == "lo":
            continue
        macFile = "/sys/class/net/%s/address" % iproc  # iproc网卡名称
        with open(macFile, "r") as fitem:
            mac = fitem.read()
            mac = mac.strip("\n")  # strip()可以去除头尾指定字符
            #在服务版 networkd 管理器才使用wifi配置 #“NetworkManager”，"networkd"
            if isWifi(iproc) and getNetplanContext() == "networkd":
                dict_mac2name_wifi[mac] = iproc
            elif not isWifi(iproc):
                dict_mac2name[mac] = iproc
    # if len(dict_mac2name_wifi) == 0:
    #     #清除之前WiFi配置 “wifi_prejson.txt”
    #     os.system("rm -f %s" % netplan_config_path_wifi)

    # 初始化列表node
    for imac in dict_mac2name:
        inode = ND_Node()
        inode.mac = imac
        inode.name = dict_mac2name[imac]
        inode.uploadConfig()
        gMName4Node[imac] = inode

    # 初始化wifi列表
    for imac_wifi in dict_mac2name_wifi:
        inode = ND_Node_Wifi()
        inode.mac = imac_wifi
        inode.name = name = dict_mac2name_wifi[imac_wifi]
        inode.uploadConfig()
        gMName4Node_wifi[imac_wifi] = inode

    # 重新构建配置文件, 处理网段重复问题(有线)
    reBuildNodes(listnode=gMName4Node)

    # 解析成netplan文件
    buildNetPlanFile(gMName4Node, gMName4Node_wifi)

    #exit(-1)
    # 执行网络计划
    dcmd = CCommand("netplan apply")
    netplan_back = dcmd.command_run()
    print("netplan apply time:" + netplan_back["RUNTIME"])

    # 回滚 更新执行结果
    finshUpdatePrejson(gMName4Node, netplan_config_path)
    finshUpdatePrejson(gMName4Node_wifi, netplan_config_path_wifi)

    print("finish")
